import { type Compiler, type Compilation, type Asset } from 'webpack';
import { FilesDependencies, Manifest, PluginOptions, AssetData, TemplateFunction } from './interfaces';
declare class ChunksWebpackPlugin {
    options: PluginOptions;
    /**
     * Instanciate the constructor
     * @param {object} options Plugin options
     */
    constructor(options: PluginOptions);
    /**
     * Apply function is automatically called by the Webpack main compiler
     * @param {Compiler} compiler The Webpack compiler variable
     */
    apply(compiler: Compiler): void;
    /**
     * Hook expose by the Webpack compiler
     * @async
     * @param {Compilation} compilation Webpack compilation
     */
    hookCallback(compilation: Compilation): Promise<void>;
    /**
     * Add assets
     * The hook is triggered by webpack
     * @async
     * @param {Compilation} compilation Webpack compilation
     * @returns {Promise<void>}
     */
    addAssets(compilation: Compilation): Promise<void>;
    /**
     * Get SVGs filtered by entrypoints
     * @param {Object} options
     * @param {Compilation} options.compilation Webpack compilation
     * @param {String} options.entryName Entrypoint name
     * @returns {Array<NormalModule>} Svgs list
     */
    getFilesDependenciesByEntrypoint({ compilation, entryName }: {
        compilation: Compilation;
        entryName: string;
    }): FilesDependencies;
    /**
     * Get the public path from Webpack configuation
     * and add slash at the end if necessary
     * @param {Compilation} compilation Webpack compilation
     * @return {String} The public path
     */
    getPublicPath(compilation: Compilation): string;
    /**
     * Get assets data
     * File path for the manifest
     * HTML tags for the generated files
     * @param {Object} options
     * @param {TemplateFunction} options.templateFunction Template function to generate HTML tags
     * @param {Asset} options.assets Asset module
     * @param {String} options.entryName Entry name
     * @param {String} options.publicPath Public path generated
     * @returns
     */
    getAssetData({ templateFunction, assets, entryName, publicPath }: {
        templateFunction: TemplateFunction;
        assets: Array<Asset>;
        entryName: string;
        publicPath: string;
    }): AssetData;
    /**
     * Create chunks manifest with Webpack compilation
     * Expose the manifest file into the assets compilation
     * The file is automatically created by the compiler
     * @async
     */
    createChunksManifestFile({ compilation, cache, eTag, manifest }: {
        compilation: Compilation;
        cache: any;
        eTag: any;
        manifest: Manifest;
    }): Promise<void>;
}
export = ChunksWebpackPlugin;
